@@ -4,20 +4,17 @@
 //! Perform operations on a card. Different states of a card are modeled by
 //! different types, such as `Open`, `User`, `Sign`, `Admin`.
 
-use anyhow::{anyhow, Result};
-
 use sequoia_openpgp::cert::amalgamation::key::ValidErasedKeyAmalgamation;
 use sequoia_openpgp::packet::key::SecretParts;
-use sequoia_openpgp::policy::Policy;
 use sequoia_openpgp::Cert;
 
 use openpgp_card::algorithm::{Algo, AlgoInfo, AlgoSimple};
 use openpgp_card::card_do::{
-    ApplicationIdentifier, ApplicationRelatedData, CardholderRelatedData,
-    ExtendedCapabilities, ExtendedLengthInfo, Fingerprint, HistoricalBytes,
-    KeyGenerationTime, PWStatusBytes, SecuritySupportTemplate, Sex,
+    ApplicationIdentifier, ApplicationRelatedData, CardholderRelatedData, ExtendedCapabilities,
+    ExtendedLengthInfo, Fingerprint, HistoricalBytes, KeyGenerationTime, Lang, PWStatusBytes,
+    SecuritySupportTemplate, Sex,
 };
-use openpgp_card::{CardApp, Error, KeySet, KeyType, Response};
+use openpgp_card::{Error, KeySet, KeyType, OpenPgpTransaction};
 
 use crate::decryptor::CardDecryptor;
 use crate::signer::CardSigner;
@@ -28,7 +25,7 @@ use openpgp_card::crypto_data::PublicKeyMaterial;
 /// Representation of an opened OpenPGP card in its base state (i.e. no
 /// passwords have been verified, default authorization applies).
 pub struct Open<'a> {
-    card_app: &'a mut CardApp,
+    opt: OpenPgpTransaction<'a>,
 
     // Cache of "application related data".
     //
@@ -47,44 +44,42 @@ pub struct Open<'a> {
 }
 
 impl<'a> Open<'a> {
-    pub fn new(card_app: &'a mut CardApp) -> Result<Self, Error> {
-        let ard = card_app.application_related_data()?;
+    pub fn new(mut ca: OpenPgpTransaction<'a>) -> Result<Self, Error> {
+        let ard = ca.application_related_data()?;
 
         Ok(Self {
-            card_app,
+            opt: ca,
             ard,
             pw1: false,
             pw1_sign: false,
             pw3: false,
         })
     }
-    pub fn feature_pinpad_verify(&self) -> bool {
-        self.card_app.feature_pinpad_verify()
+
+    pub fn feature_pinpad_verify(&mut self) -> bool {
+        self.opt.feature_pinpad_verify()
     }
 
-    pub fn feature_pinpad_modify(&self) -> bool {
-        self.card_app.feature_pinpad_modify()
+    pub fn feature_pinpad_modify(&mut self) -> bool {
+        self.opt.feature_pinpad_modify()
     }
 
     pub fn verify_user(&mut self, pin: &str) -> Result<(), Error> {
-        let _ = self.card_app.verify_pw1(pin)?;
+        let _ = self.opt.verify_pw1_user(pin.as_bytes())?;
         self.pw1 = true;
         Ok(())
     }
 
-    pub fn verify_user_pinpad(
-        &mut self,
-        prompt: &dyn Fn(),
-    ) -> Result<(), Error> {
+    pub fn verify_user_pinpad(&mut self, prompt: &dyn Fn()) -> Result<(), Error> {
         prompt();
 
-        let _ = self.card_app.verify_pw1_pinpad()?;
+        let _ = self.opt.verify_pw1_user_pinpad()?;
         self.pw1 = true;
         Ok(())
     }
 
     pub fn verify_user_for_signing(&mut self, pin: &str) -> Result<(), Error> {
-        let _ = self.card_app.verify_pw1_for_signing(pin)?;
+        let _ = self.opt.verify_pw1_sign(pin.as_bytes())?;
 
         // FIXME: depending on card mode, pw1_sign is only usable once
 
@@ -92,13 +87,10 @@ impl<'a> Open<'a> {
         Ok(())
     }
 
-    pub fn verify_user_for_signing_pinpad(
-        &mut self,
-        prompt: &dyn Fn(),
-    ) -> Result<(), Error> {
+    pub fn verify_user_for_signing_pinpad(&mut self, prompt: &dyn Fn()) -> Result<(), Error> {
         prompt();
 
-        let _ = self.card_app.verify_pw1_for_signing_pinpad()?;
+        let _ = self.opt.verify_pw1_sign_pinpad()?;
 
         // FIXME: depending on card mode, pw1_sign is only usable once
 
@@ -107,18 +99,15 @@ impl<'a> Open<'a> {
     }
 
     pub fn verify_admin(&mut self, pin: &str) -> Result<(), Error> {
-        let _ = self.card_app.verify_pw3(pin)?;
+        let _ = self.opt.verify_pw3(pin.as_bytes())?;
         self.pw3 = true;
         Ok(())
     }
 
-    pub fn verify_admin_pinpad(
-        &mut self,
-        prompt: &dyn Fn(),
-    ) -> Result<(), Error> {
+    pub fn verify_admin_pinpad(&mut self, prompt: &dyn Fn()) -> Result<(), Error> {
         prompt();
 
-        let _ = self.card_app.verify_pw3_pinpad()?;
+        let _ = self.opt.verify_pw3_pinpad()?;
         self.pw3 = true;
         Ok(())
     }
@@ -126,60 +115,42 @@ impl<'a> Open<'a> {
     /// Ask the card if the user password has been successfully verified.
     ///
     /// NOTE: on some cards this functionality seems broken.
-    pub fn check_user_verified(&mut self) -> Result<Response, Error> {
-        self.card_app.check_pw1()
+    pub fn check_user_verified(&mut self) -> Result<(), Error> {
+        self.opt.check_pw1_user()
     }
 
     /// Ask the card if the admin password has been successfully verified.
     ///
     /// NOTE: on some cards this functionality seems broken.
-    pub fn check_admin_verified(&mut self) -> Result<Response, Error> {
-        self.card_app.check_pw3()
+    pub fn check_admin_verified(&mut self) -> Result<(), Error> {
+        self.opt.check_pw3()
     }
 
-    pub fn change_user_pin(
-        &mut self,
-        old: &str,
-        new: &str,
-    ) -> Result<Response, Error> {
-        self.card_app.change_pw1(old, new)
+    pub fn change_user_pin(&mut self, old: &str, new: &str) -> Result<(), Error> {
+        self.opt.change_pw1(old.as_bytes(), new.as_bytes())
     }
 
-    pub fn change_user_pin_pinpad(
-        &mut self,
-        prompt: &dyn Fn(),
-    ) -> Result<Response, Error> {
+    pub fn change_user_pin_pinpad(&mut self, prompt: &dyn Fn()) -> Result<(), Error> {
         prompt();
-        self.card_app.change_pw1_pinpad()
+        self.opt.change_pw1_pinpad()
     }
 
-    pub fn reset_user_pin(
-        &mut self,
-        rst: &str,
-        new: &str,
-    ) -> Result<Response, Error> {
-        self.card_app
-            .reset_retry_counter_pw1(new.into(), Some(rst.into()))
+    pub fn reset_user_pin(&mut self, rst: &str, new: &str) -> Result<(), Error> {
+        self.opt
+            .reset_retry_counter_pw1(new.as_bytes(), Some(rst.as_bytes()))
     }
 
-    pub fn change_admin_pin(
-        &mut self,
-        old: &str,
-        new: &str,
-    ) -> Result<Response, Error> {
-        self.card_app.change_pw3(old, new)
+    pub fn change_admin_pin(&mut self, old: &str, new: &str) -> Result<(), Error> {
+        self.opt.change_pw3(old.as_bytes(), new.as_bytes())
     }
 
-    pub fn change_admin_pin_pinpad(
-        &mut self,
-        prompt: &dyn Fn(),
-    ) -> Result<Response, Error> {
+    pub fn change_admin_pin_pinpad(&mut self, prompt: &dyn Fn()) -> Result<(), Error> {
         prompt();
-        self.card_app.change_pw3_pinpad()
+        self.opt.change_pw3_pinpad()
     }
 
     /// Get a view of the card authenticated for "User" commands.
-    pub fn user_card<'b>(&'a mut self) -> Option<User<'a, 'b>> {
+    pub fn user_card<'b>(&'b mut self) -> Option<User<'a, 'b>> {
         if self.pw1 {
             Some(User { oc: self })
         } else {
@@ -207,9 +178,7 @@ impl<'a> Open<'a> {
 
     // --- application data ---
 
-    pub fn application_identifier(
-        &self,
-    ) -> Result<ApplicationIdentifier, Error> {
+    pub fn application_identifier(&self) -> Result<ApplicationIdentifier, Error> {
         self.ard.application_id()
     }
 
@@ -217,9 +186,7 @@ impl<'a> Open<'a> {
         self.ard.historical_bytes()
     }
 
-    pub fn extended_length_information(
-        &self,
-    ) -> Result<Option<ExtendedLengthInfo>> {
+    pub fn extended_length_information(&self) -> Result<Option<ExtendedLengthInfo>, Error> {
         self.ard.extended_length_information()
     }
 
@@ -233,18 +200,16 @@ impl<'a> Open<'a> {
         unimplemented!()
     }
 
-    pub fn extended_capabilities(
-        &self,
-    ) -> Result<ExtendedCapabilities, Error> {
+    pub fn extended_capabilities(&self) -> Result<ExtendedCapabilities, Error> {
         self.ard.extended_capabilities()
     }
 
-    pub fn algorithm_attributes(&self, key_type: KeyType) -> Result<Algo> {
+    pub fn algorithm_attributes(&self, key_type: KeyType) -> Result<Algo, Error> {
         self.ard.algorithm_attributes(key_type)
     }
 
     /// PW status Bytes
-    pub fn pw_status_bytes(&self) -> Result<PWStatusBytes> {
+    pub fn pw_status_bytes(&self) -> Result<PWStatusBytes, Error> {
         self.ard.pw_status_bytes()
     }
 
@@ -257,9 +222,7 @@ impl<'a> Open<'a> {
         unimplemented!()
     }
 
-    pub fn key_generation_times(
-        &self,
-    ) -> Result<KeySet<KeyGenerationTime>, Error> {
+    pub fn key_generation_times(&self) -> Result<KeySet<KeyGenerationTime>, Error> {
         self.ard.key_generation_times()
     }
 
@@ -294,26 +257,22 @@ impl<'a> Open<'a> {
 
     // --- URL (5f50) ---
 
-    pub fn url(&mut self) -> Result<String> {
-        self.card_app.url()
+    pub fn url(&mut self) -> Result<String, Error> {
+        Ok(String::from_utf8_lossy(&self.opt.url()?).to_string())
     }
 
     // --- cardholder related data (65) ---
-    pub fn cardholder_related_data(
-        &mut self,
-    ) -> Result<CardholderRelatedData> {
-        self.card_app.cardholder_related_data()
+    pub fn cardholder_related_data(&mut self) -> Result<CardholderRelatedData, Error> {
+        self.opt.cardholder_related_data()
     }
 
     // --- security support template (7a) ---
-    pub fn security_support_template(
-        &mut self,
-    ) -> Result<SecuritySupportTemplate> {
-        self.card_app.security_support_template()
+    pub fn security_support_template(&mut self) -> Result<SecuritySupportTemplate, Error> {
+        self.opt.security_support_template()
     }
 
     // DO "Algorithm Information" (0xFA)
-    pub fn algorithm_information(&mut self) -> Result<Option<AlgoInfo>> {
+    pub fn algorithm_information(&mut self) -> Result<Option<AlgoInfo>, Error> {
         // The DO "Algorithm Information" (Tag FA) shall be present if
         // Algorithm attributes can be changed
         let ec = self.extended_capabilities()?;
@@ -323,28 +282,25 @@ impl<'a> Open<'a> {
             return Ok(None);
         }
 
-        self.card_app.algorithm_information()
+        self.opt.algorithm_information()
     }
 
     /// Firmware Version, YubiKey specific (?)
-    pub fn firmware_version(&mut self) -> Result<Vec<u8>> {
-        self.card_app.firmware_version()
+    pub fn firmware_version(&mut self) -> Result<Vec<u8>, Error> {
+        self.opt.firmware_version()
     }
 
     // ----------
 
-    pub fn public_key(
-        &mut self,
-        key_type: KeyType,
-    ) -> Result<PublicKeyMaterial> {
-        self.card_app.public_key(key_type).map_err(|e| e.into())
+    pub fn public_key(&mut self, key_type: KeyType) -> Result<PublicKeyMaterial, Error> {
+        self.opt.public_key(key_type).map_err(|e| e.into())
     }
 
     // ----------
 
     /// Delete all state on this OpenPGP card
-    pub fn factory_reset(&mut self) -> Result<()> {
-        self.card_app.factory_reset()
+    pub fn factory_reset(&mut self) -> Result<(), Error> {
+        self.opt.factory_reset()
     }
 }
 
@@ -354,13 +310,9 @@ pub struct User<'app, 'open> {
     oc: &'open mut Open<'app>,
 }
 
-impl User<'_, '_> {
-    pub fn decryptor(
-        &mut self,
-        cert: &Cert,
-        policy: &dyn Policy,
-    ) -> Result<CardDecryptor, Error> {
-        CardDecryptor::new(&mut self.oc.card_app, cert, policy)
+impl<'app, 'open> User<'app, 'open> {
+    pub fn decryptor(&mut self, cert: &Cert) -> Result<CardDecryptor<'_, 'app>, Error> {
+        CardDecryptor::new(&mut self.oc.opt, cert)
     }
 }
 
@@ -370,23 +322,19 @@ pub struct Sign<'app, 'open> {
     oc: &'open mut Open<'app>,
 }
 
-impl Sign<'_, '_> {
-    pub fn signer(
-        &mut self,
-        cert: &Cert,
-        policy: &dyn Policy,
-    ) -> std::result::Result<CardSigner, Error> {
+impl<'app, 'open> Sign<'app, 'open> {
+    pub fn signer(&mut self, cert: &Cert) -> std::result::Result<CardSigner<'_, 'app>, Error> {
         // FIXME: depending on the setting in "PW1 Status byte", only one
         // signature can be made after verification for signing
 
-        CardSigner::new(&mut self.oc.card_app, cert, policy)
+        CardSigner::new(&mut self.oc.opt, cert)
     }
 
-    pub fn signer_from_pubkey(&mut self, pubkey: PublicKey) -> CardSigner {
+    pub fn signer_from_pubkey(&mut self, pubkey: PublicKey) -> CardSigner<'_, 'app> {
         // FIXME: depending on the setting in "PW1 Status byte", only one
         // signature can be made after verification for signing
 
-        CardSigner::with_pubkey(&mut self.oc.card_app, pubkey)
+        CardSigner::with_pubkey(&mut self.oc.opt, pubkey)
     }
 }
 
@@ -402,61 +350,57 @@ impl<'app, 'open> Admin<'app, 'open> {
 }
 
 impl Admin<'_, '_> {
-    pub fn set_name(&mut self, name: &str) -> Result<Response, Error> {
+    pub fn set_name(&mut self, name: &str) -> Result<(), Error> {
         if name.len() >= 40 {
-            return Err(anyhow!("name too long").into());
+            return Err(Error::InternalError("name too long".into()));
         }
 
         // All chars must be in ASCII7
         if name.chars().any(|c| !c.is_ascii()) {
-            return Err(anyhow!("Invalid char in name").into());
+            return Err(Error::InternalError("Invalid char in name".into()));
         };
 
-        self.oc.card_app.set_name(name)
+        self.oc.opt.set_name(name.as_bytes())
     }
 
-    pub fn set_lang(&mut self, lang: &str) -> Result<Response, Error> {
+    pub fn set_lang(&mut self, lang: &[Lang]) -> Result<(), Error> {
         if lang.len() > 8 {
-            return Err(anyhow!("lang too long").into());
+            return Err(Error::InternalError("lang too long".into()));
         }
 
-        self.oc.card_app.set_lang(lang)
+        self.oc.opt.set_lang(lang)
     }
 
-    pub fn set_sex(&mut self, sex: Sex) -> Result<Response, Error> {
-        self.oc.card_app.set_sex(sex)
+    pub fn set_sex(&mut self, sex: Sex) -> Result<(), Error> {
+        self.oc.opt.set_sex(sex)
     }
 
-    pub fn set_url(&mut self, url: &str) -> Result<Response, Error> {
+    pub fn set_url(&mut self, url: &str) -> Result<(), Error> {
         if url.chars().any(|c| !c.is_ascii()) {
-            return Err(anyhow!("Invalid char in url").into());
+            return Err(Error::InternalError("Invalid char in url".into()));
         }
 
         // Check for max len
         let ec = self.oc.extended_capabilities()?;
 
-        if ec.max_len_special_do() == None
-            || url.len() <= ec.max_len_special_do().unwrap() as usize
+        if ec.max_len_special_do() == None || url.len() <= ec.max_len_special_do().unwrap() as usize
         {
             // If we don't know the max length for URL ("special DO"),
             // or if it's within the acceptable length:
             // send the url update to the card.
 
-            self.oc.card_app.set_url(url)
+            self.oc.opt.set_url(url.as_bytes())
         } else {
-            Err(anyhow!("URL too long").into())
+            Err(Error::InternalError("URL too long".into()))
         }
     }
 
-    pub fn set_resetting_code(
-        &mut self,
-        pin: &str,
-    ) -> Result<Response, Error> {
-        self.oc.card_app.set_resetting_code(pin.into())
+    pub fn set_resetting_code(&mut self, pin: &str) -> Result<(), Error> {
+        self.oc.opt.set_resetting_code(pin.as_bytes())
     }
 
-    pub fn reset_user_pin(&mut self, new: &str) -> Result<Response, Error> {
-        self.oc.card_app.reset_retry_counter_pw1(new.into(), None)
+    pub fn reset_user_pin(&mut self, new: &str) -> Result<(), Error> {
+        self.oc.opt.reset_retry_counter_pw1(new.as_bytes(), None)
     }
 
     /// Upload a ValidErasedKeyAmalgamation to the card as a specific KeyType.
@@ -469,7 +413,7 @@ impl Admin<'_, '_> {
         password: Option<String>,
     ) -> Result<(), Error> {
         let key = vka_as_uploadable_key(vka, password);
-        self.oc.card_app.key_import(key, key_type)
+        self.oc.opt.key_import(key, key_type)
     }
 
     pub fn generate_key_simple(
@@ -478,16 +422,14 @@ impl Admin<'_, '_> {
         algo: Option<AlgoSimple>,
     ) -> Result<(PublicKeyMaterial, KeyGenerationTime), Error> {
         match algo {
-            Some(algo) => self.oc.card_app.generate_key_simple(
-                public_to_fingerprint,
-                key_type,
-                algo,
-            ),
-            None => self.oc.card_app.generate_key(
-                public_to_fingerprint,
-                key_type,
-                None,
-            ),
+            Some(algo) => self
+                .oc
+                .opt
+                .generate_key_simple(public_to_fingerprint, key_type, algo),
+            None => self
+                .oc
+                .opt
+                .generate_key(public_to_fingerprint, key_type, None),
         }
     }
 }
