@@ -10,8 +10,8 @@ use sequoia_openpgp::policy::StandardPolicy;
 use sequoia_openpgp::Cert;
 
 use openpgp_card::card_do::Sex;
-use openpgp_card::KeyType;
-use openpgp_card_pcsc::PcscClient;
+use openpgp_card::{KeyType, OpenPgp};
+use openpgp_card_pcsc::PcscBackend;
 
 use openpgp_card_sequoia::card::Open;
 use openpgp_card_sequoia::sq_util;
@@ -35,8 +35,10 @@ fn main() -> Result<(), Box<dyn Error>> {
     let test_card_ident = env::var("TEST_CARD_IDENT");
 
     if let Ok(test_card_ident) = test_card_ident {
-        let mut card = PcscClient::open_by_ident(&test_card_ident)?;
-        let mut open = Open::new(&mut card)?;
+        let mut card = PcscBackend::open_by_ident(&test_card_ident, None)?;
+        let mut pgp = OpenPgp::new(&mut card);
+
+        let mut open = Open::new(pgp.transaction()?)?;
 
         // card metadata
 
@@ -109,7 +111,7 @@ fn main() -> Result<(), Box<dyn Error>> {
         let res = admin.set_sex(Sex::NotApplicable)?;
         println!("set sex {:x?}", res);
 
-        let res = admin.set_lang("en")?;
+        let res = admin.set_lang(&[['e', 'n'].into()])?;
         println!("set lang {:x?}", res);
 
         let res = admin.set_url("https://keys.openpgp.org")?;
@@ -118,23 +120,17 @@ fn main() -> Result<(), Box<dyn Error>> {
         let cert = Cert::from_file(TEST_KEY_PATH)?;
         let p = StandardPolicy::new();
 
-        if let Some(vka) =
-            sq_util::subkey_by_type(&cert, &p, KeyType::Signing)?
-        {
+        if let Some(vka) = sq_util::subkey_by_type(&cert, &p, KeyType::Signing)? {
             println!("Upload signing key");
             admin.upload_key(vka, KeyType::Signing, None)?;
         }
 
-        if let Some(vka) =
-            sq_util::subkey_by_type(&cert, &p, KeyType::Decryption)?
-        {
+        if let Some(vka) = sq_util::subkey_by_type(&cert, &p, KeyType::Decryption)? {
             println!("Upload decryption key");
             admin.upload_key(vka, KeyType::Decryption, None)?;
         }
 
-        if let Some(vka) =
-            sq_util::subkey_by_type(&cert, &p, KeyType::Authentication)?
-        {
+        if let Some(vka) = sq_util::subkey_by_type(&cert, &p, KeyType::Authentication)? {
             println!("Upload auth key");
             admin.upload_key(vka, KeyType::Authentication, None)?;
         }
@@ -144,8 +140,10 @@ fn main() -> Result<(), Box<dyn Error>> {
         // -----------------------------
         //  Open fresh Card for decrypt
         // -----------------------------
-        let mut card = PcscClient::open_by_ident(&test_card_ident)?;
-        let mut open = Open::new(&mut card)?;
+        let mut card = PcscBackend::open_by_ident(&test_card_ident, None)?;
+        let mut pgp = OpenPgp::new(&mut card);
+
+        let mut open = Open::new(pgp.transaction()?)?;
 
         // Check that we're still using the expected card
         let app_id = open.application_identifier()?;
@@ -166,13 +164,12 @@ fn main() -> Result<(), Box<dyn Error>> {
             .expect("We just validated, this should not fail");
 
         let cert = Cert::from_file(TEST_KEY_PATH)?;
-        let msg = std::fs::read_to_string(TEST_ENC_MSG)
-            .expect("Unable to read file");
+        let msg = std::fs::read_to_string(TEST_ENC_MSG).expect("Unable to read file");
 
         println!("Encrypted message:\n{}", msg);
 
         let sp = StandardPolicy::new();
-        let d = user.decryptor(&cert, &sp)?;
+        let d = user.decryptor(&cert)?;
         let res = sq_util::decryption_helper(d, msg.into_bytes(), &sp)?;
 
         let plain = String::from_utf8_lossy(&res);
@@ -183,8 +180,10 @@ fn main() -> Result<(), Box<dyn Error>> {
         // -----------------------------
         //  Open fresh Card for signing
         // -----------------------------
-        let mut card = PcscClient::open_by_ident(&test_card_ident)?;
-        let mut open = Open::new(&mut card)?;
+        let mut card = PcscBackend::open_by_ident(&test_card_ident, None)?;
+        let mut pgp = OpenPgp::new(&mut card);
+
+        let mut open = Open::new(pgp.transaction()?)?;
 
         // Sign
         open.verify_user_for_signing("123456")?;
@@ -197,7 +196,7 @@ fn main() -> Result<(), Box<dyn Error>> {
 
         let text = "Hello world, I am signed.";
 
-        let signer = sign.signer(&cert, &StandardPolicy::new())?;
+        let signer = sign.signer(&cert)?;
         let sig = sq_util::sign_helper(signer, &mut text.as_bytes())?;
 
         println!("Signature from card:\n{}", sig)
@@ -213,8 +212,11 @@ fn main() -> Result<(), Box<dyn Error>> {
 
         println!("The following OpenPGP cards are connected to your system:");
 
-        for mut card in PcscClient::cards()? {
-            let open = Open::new(&mut card)?;
+        for mut card in PcscBackend::cards(None)? {
+            let mut pgp = OpenPgp::new(&mut card);
+
+            let open = Open::new(pgp.transaction()?)?;
+
             println!(" {}", open.application_identifier()?.ident());
         }
     }
