@@ -7,34 +7,32 @@ use anyhow::anyhow;
 
 use openpgp::crypto;
 use openpgp::crypto::mpi;
-use openpgp::policy::Policy;
 use openpgp::types::{Curve, PublicKeyAlgorithm};
 use sequoia_openpgp as openpgp;
 
 use openpgp_card::crypto_data::Hash;
-use openpgp_card::{CardApp, Error};
+use openpgp_card::{Error, OpenPgpTransaction};
 
 use crate::sq_util;
 use crate::PublicKey;
 
-pub struct CardSigner<'a> {
+pub struct CardSigner<'a, 'app> {
     /// The OpenPGP card (authenticated to allow signing operations)
-    ca: &'a mut CardApp,
+    ca: &'a mut OpenPgpTransaction<'app>,
 
     /// The matching public key for the card's signing key
     public: PublicKey,
 }
 
-impl<'a> CardSigner<'a> {
+impl<'a, 'app> CardSigner<'a, 'app> {
     /// Try to create a CardSigner.
     ///
     /// An Error is returned if no match between the card's signing
     /// key and a (sub)key of `cert` can be made.
     pub fn new(
-        ca: &'a mut CardApp,
+        ca: &'a mut OpenPgpTransaction<'app>,
         cert: &openpgp::Cert,
-        policy: &dyn Policy,
-    ) -> Result<CardSigner<'a>, Error> {
+    ) -> Result<CardSigner<'a, 'app>, Error> {
         // Get the fingerprint for the signing key from the card.
         let ard = ca.application_related_data()?;
         let fps = ard.fingerprints()?;
@@ -44,40 +42,31 @@ impl<'a> CardSigner<'a> {
             // Transform into Sequoia Fingerprint
             let fp = openpgp::Fingerprint::from_bytes(fp.as_bytes());
 
-            if let Some(vk) =
-                sq_util::get_subkey_by_fingerprint(cert, policy, &fp, true)?
-            {
-                if vk.for_signing() {
-                    let key = vk.key().clone();
-                    Ok(Self::with_pubkey(ca, key))
-                } else {
-                    Err(Error::InternalError(anyhow!(
-                        "(Sub)key {} in the cert isn't signing capable",
-                        fp
-                    )))
-                }
+            if let Some(eka) = sq_util::get_subkey_by_fingerprint(cert, &fp)? {
+                let key = eka.key().clone();
+                Ok(Self::with_pubkey(ca, key))
             } else {
-                Err(Error::InternalError(anyhow!(
+                Err(Error::InternalError(format!(
                     "Failed to find (sub)key {} in cert",
                     fp
                 )))
             }
         } else {
-            Err(Error::InternalError(anyhow!(
+            Err(Error::InternalError(format!(
                 "Failed to get the signing key's Fingerprint from the card"
             )))
         }
     }
 
     pub(crate) fn with_pubkey(
-        ca: &'a mut CardApp,
+        ca: &'a mut OpenPgpTransaction<'app>,
         public: PublicKey,
-    ) -> CardSigner<'a> {
+    ) -> CardSigner<'a, 'app> {
         CardSigner { ca, public }
     }
 }
 
-impl<'a> crypto::Signer for CardSigner<'a> {
+impl<'a, 'app> crypto::Signer for CardSigner<'a, 'app> {
     fn public(&self) -> &PublicKey {
         &self.public
     }
@@ -97,10 +86,7 @@ impl<'a> crypto::Signer for CardSigner<'a> {
         match (self.public.pk_algo(), self.public.mpis()) {
             #[allow(deprecated)]
             (PublicKeyAlgorithm::RSASign, mpi::PublicKey::RSA { .. })
-            | (
-                PublicKeyAlgorithm::RSAEncryptSign,
-                mpi::PublicKey::RSA { .. },
-            ) => {
+            | (PublicKeyAlgorithm::RSAEncryptSign, mpi::PublicKey::RSA { .. }) => {
                 let hash = match hash_algo {
                     openpgp::types::HashAlgorithm::SHA256 => Hash::SHA256(
                         digest
@@ -140,10 +126,7 @@ impl<'a> crypto::Signer for CardSigner<'a> {
 
                 Ok(mpi::Signature::EdDSA { r, s })
             }
-            (
-                PublicKeyAlgorithm::ECDSA,
-                mpi::PublicKey::ECDSA { curve, .. },
-            ) => {
+            (PublicKeyAlgorithm::ECDSA, mpi::PublicKey::ECDSA { curve, .. }) => {
                 let hash = match curve {
                     Curve::NistP256 => Hash::ECDSA(&digest[..32]),
                     Curve::NistP384 => Hash::ECDSA(&digest[..48]),
