@@ -19,7 +19,6 @@ use openpgp::packet::{
     Key, UserID,
 };
 use openpgp::parse::{stream::DecryptorBuilder, Parse};
-use openpgp::policy::Policy;
 use openpgp::serialize::stream::{Message, Signer};
 use openpgp::types::{KeyFlags, PublicKeyAlgorithm, SignatureType, Timestamp};
 use openpgp::{Cert, Packet};
@@ -28,11 +27,12 @@ use sequoia_openpgp as openpgp;
 use openpgp_card::algorithm::{Algo, Curve};
 use openpgp_card::card_do::{Fingerprint, KeyGenerationTime};
 use openpgp_card::crypto_data::{CardUploadableKey, PublicKeyMaterial};
-use openpgp_card::{CardApp, Error, KeyType};
+use openpgp_card::{Error, KeyType, OpenPgpTransaction};
 
 use crate::card::Open;
 use crate::privkey::SequoiaKey;
 use crate::{decryptor, signer, PublicKey};
+use sequoia_openpgp::policy::Policy;
 
 /// Create a Cert from the three subkeys on a card.
 /// (Calling this multiple times will result in different Certs!)
@@ -66,15 +66,14 @@ pub fn make_cert<'a, 'app>(
 
         // 3) make binding, sign with card -> add
         {
-            let signing_builder =
-                SignatureBuilder::new(SignatureType::SubkeyBinding)
-                    .set_signature_creation_time(SystemTime::now())?
-                    .set_key_validity_period(std::time::Duration::new(0, 0))?
-                    .set_key_flags(
-                        KeyFlags::empty()
-                            .set_storage_encryption()
-                            .set_transport_encryption(),
-                    )?;
+            let signing_builder = SignatureBuilder::new(SignatureType::SubkeyBinding)
+                .set_signature_creation_time(SystemTime::now())?
+                .set_key_validity_period(std::time::Duration::new(0, 0))?
+                .set_key_flags(
+                    KeyFlags::empty()
+                        .set_storage_encryption()
+                        .set_transport_encryption(),
+                )?;
 
             // Allow signing on the card
             if let Some(pw1) = pw1.clone() {
@@ -102,11 +101,10 @@ pub fn make_cert<'a, 'app>(
 
         // 5) make, sign binding -> add
         {
-            let signing_builder =
-                SignatureBuilder::new(SignatureType::SubkeyBinding)
-                    .set_signature_creation_time(SystemTime::now())?
-                    .set_key_validity_period(std::time::Duration::new(0, 0))?
-                    .set_key_flags(KeyFlags::empty().set_authentication())?;
+            let signing_builder = SignatureBuilder::new(SignatureType::SubkeyBinding)
+                .set_signature_creation_time(SystemTime::now())?
+                .set_key_validity_period(std::time::Duration::new(0, 0))?
+                .set_key_flags(KeyFlags::empty().set_authentication())?;
 
             // Allow signing on the card
             if let Some(pw1) = pw1.clone() {
@@ -137,21 +135,19 @@ pub fn make_cert<'a, 'app>(
 
     // FIXME: accept email as argument?!
 
-    let uid: UserID =
-        cardholder.name().expect("expecting name on card").into();
+    let uid: UserID = cardholder.name().expect("expecting name on card").into();
 
     pp.push(uid.clone().into());
 
     // 7) make, sign binding -> add
     {
-        let signing_builder =
-            SignatureBuilder::new(SignatureType::PositiveCertification)
-                .set_signature_creation_time(SystemTime::now())?
-                .set_key_validity_period(std::time::Duration::new(0, 0))?
-                .set_key_flags(
-                    // Flags for primary key
-                    KeyFlags::empty().set_signing().set_certification(),
-                )?;
+        let signing_builder = SignatureBuilder::new(SignatureType::PositiveCertification)
+            .set_signature_creation_time(SystemTime::now())?
+            .set_key_validity_period(std::time::Duration::new(0, 0))?
+            .set_key_flags(
+                // Flags for primary key
+                KeyFlags::empty().set_signing().set_certification(),
+            )?;
 
         // Allow signing on the card
         if let Some(pw1) = pw1 {
@@ -167,8 +163,7 @@ pub fn make_cert<'a, 'app>(
             // Temporary version of the cert
             let cert = Cert::try_from(pp.clone())?;
 
-            let signing_bsig: Packet =
-                uid.bind(&mut card_signer, &cert, signing_builder)?.into();
+            let signing_bsig: Packet = uid.bind(&mut card_signer, &cert, signing_builder)?.into();
 
             pp.push(signing_bsig);
         }
@@ -182,12 +177,14 @@ pub fn public_key_material_to_key(
     pkm: &PublicKeyMaterial,
     key_type: KeyType,
     time: KeyGenerationTime,
-) -> Result<PublicKey> {
+) -> Result<PublicKey, Error> {
     let time = Timestamp::from(time.get()).into();
 
     match pkm {
         PublicKeyMaterial::R(rsa) => {
-            let k4 = Key4::import_public_rsa(rsa.v(), rsa.n(), Some(time))?;
+            let k4 = Key4::import_public_rsa(rsa.v(), rsa.n(), Some(time)).map_err(|e| {
+                Error::InternalError(format!("sequoia Key4::import_public_rsa failed: {:?}", e))
+            })?;
 
             Ok(k4.into())
         }
@@ -208,7 +205,12 @@ pub fn public_key_material_to_key(
                         if algo_ecc.curve() == Curve::Ed25519 {
                             // EdDSA
                             let k4 =
-                                Key4::import_public_ed25519(ecc.data(), time)?;
+                                Key4::import_public_ed25519(ecc.data(), time).map_err(|e| {
+                                    Error::InternalError(format!(
+                                        "sequoia Key4::import_public_ed25519 failed: {:?}",
+                                        e
+                                    ))
+                                })?;
 
                             Ok(Key::from(k4))
                         } else {
@@ -220,20 +222,30 @@ pub fn public_key_material_to_key(
                                     curve,
                                     q: mpi::MPI::new(ecc.data()),
                                 },
-                            )?;
+                            )
+                            .map_err(|e| {
+                                Error::InternalError(format!(
+                                    "sequoia Key4::new for ECDSA failed: {:?}",
+                                    e
+                                ))
+                            })?;
 
                             Ok(k4.into())
                         }
                     }
                     KeyType::Decryption => {
                         if algo_ecc.curve() == Curve::Cv25519 {
+                            // FIXME: not setting `hash` and `sym` is not
+                            // ok when a cert already exists
+
                             // EdDSA
-                            let k4 = Key4::import_public_cv25519(
-                                ecc.data(),
-                                None,
-                                None,
-                                time,
-                            )?;
+                            let k4 = Key4::import_public_cv25519(ecc.data(), None, None, time)
+                                .map_err(|e| {
+                                    Error::InternalError(format!(
+                                        "sequoia Key4::import_public_cv25519 failed: {:?}",
+                                        e
+                                    ))
+                                })?;
 
                             Ok(k4.into())
                         } else {
@@ -250,7 +262,13 @@ pub fn public_key_material_to_key(
                                     hash: Default::default(),
                                     sym: Default::default(),
                                 },
-                            )?;
+                            )
+                            .map_err(|e| {
+                                Error::InternalError(format!(
+                                    "sequoia Key4::new for ECDH failed: {:?}",
+                                    e
+                                ))
+                            })?;
 
                             Ok(k4.into())
                         }
@@ -291,14 +309,13 @@ pub fn vka_as_uploadable_key(
 
 /// FIXME: this fn is used in card_functionality, but should be removed
 pub fn sign(
-    ca: &mut CardApp,
+    card_tx: &'_ mut OpenPgpTransaction<'_>,
     cert: &Cert,
     input: &mut dyn io::Read,
-    p: &dyn Policy,
 ) -> Result<String> {
     let mut armorer = armor::Writer::new(vec![], armor::Kind::Signature)?;
     {
-        let s = signer::CardSigner::new(ca, cert, p)?;
+        let s = signer::CardSigner::new(card_tx, cert)?;
 
         let message = Message::new(&mut armorer);
         let mut message = Signer::new(message, s).detached().build()?;
@@ -316,7 +333,7 @@ pub fn sign(
 
 /// FIXME: this fn is used in card_functionality, but should be removed
 pub fn decrypt(
-    ca: &mut CardApp,
+    card_tx: &'_ mut OpenPgpTransaction<'_>,
     cert: &Cert,
     msg: Vec<u8>,
     p: &dyn Policy,
@@ -325,7 +342,7 @@ pub fn decrypt(
     {
         let reader = io::BufReader::new(&msg[..]);
 
-        let d = decryptor::CardDecryptor::new(ca, cert, p)?;
+        let d = decryptor::CardDecryptor::new(card_tx, cert)?;
 
         let db = DecryptorBuilder::from_reader(reader)?;
         let mut decryptor = db.with_policy(p, None, d)?;
